# -*- coding: utf-8 -*-
"""
Copyright (c) 2021 気象データ×IT勉強会，吉野　純，岐阜大学工学部附属応用気象研究センター

Permission is hereby granted, free of charge, to any person obtaining a copy 
of this software and associated documentation files (the "Software"), to deal 
in the Software without restriction, including without limitation the rights 
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
copies of the Software, and to permit persons to whom the Software is 
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
SOFTWARE.
"""
# numpyをnpという別名でインポートします．
import numpy as np
# matplotlibをpltという別名でインポートします．
import matplotlib.pyplot as plt
# Seabornをインポートします．
import seaborn as sns
# datetimeは日時データを処理する際に便利なメソッドです．インポートします．
from datetime import datetime
# pandasをpdという別名でインポートします．
import pandas as pd
# matplotlibで時系列図を作成するときには以下をインポートします
from pandas.plotting import register_matplotlib_converters
# これを登録しておきます．
register_matplotlib_converters()
# sklearn(scikit-learn)の前処理のためのライブラリをインポートします
from sklearn import preprocessing
# sklearn(scikit-learn)の線形重回帰モデルをインポートします．
from sklearn import linear_model
# sklearn(scikit-learn)のニューラルネットワークのライブラリをインポートします
from sklearn import neural_network
# sklearn(scikit-learn)のサポートベクターマシンのライブラリをインポートします
from sklearn.svm import SVR
import pickle
#%precision 3
#%matplotlib inline

#def 1.############################################
# 気象庁アメダスの気温の時系列データを読み込んで，
# データフレームに割り当てる関数
def readamedas(filename,skipline): 
    # pandasのread_csvというメソッドでcsvファイルを読み込みます．
    # 引数として，
    # [0]入力ファイル名
    # [1]エンコーディング
    # [2]読み飛ばす行数，
    # [3]column名
    # [4]datetime型で読み込むcolumn名
    # [5]indexとするcolumn名
    # を与える．
    df=pd.read_csv(
       filename,
       encoding='Shift_JIS',
       skiprows=skipline,
       names=['date_time','T','dummy1','dummy2','Td','dummy3','dummy4'],
       parse_dates={'datetime':['date_time']}, 
       index_col='datetime'
       )
    return df

#def 2.############################################
# 東京電力の電力消費量の時系列データを読み込んで，
# DataFrameに割り当てる関数
def readtepco(filename,skipline): 
    # pandasのread_csvというメソッドでcsvファイルを読み込みます．
    # 引数として，
    # [0]入力ファイル名
    # [1]エンコーディング
    # [2]読み飛ばす行数，
    # [3]column名
    # [4]datetime型で読み込むcolumn名
    # [5]indexとするcolumn名
    # を与える．
    df=pd.read_csv(
       filename,
       encoding='Shift_JIS',
       skiprows=skipline,
       names=['date','time','POWER'],
       parse_dates={'datetime':['date','time']},
       index_col='datetime'
       )
    return df

#def 3.############################################
# 2つのデータフレームを結合し，欠測値を除去し，予測モデルの説明変数を
# 列として追加する関数．ここでは，湿数，時刻，日にち，曜日，月，過去の発電量
# を抽出・計算し，データフレームdfに連結している．
def dataprocess(x_data, y_data):
    # 2つのデータフレーム（x_dataとy_data）を結合してデータフレームdfとします．
    df=pd.merge(y_data, x_data, how="outer", left_index=True, right_index=True)
    # NaNがある行を取り除く
    df=df.dropna(how='any')
    # T-Tdにより湿数を計算して、T-Tdの列を作成し、もう使用しないTdは削除する．
    df['T-Td']=df['T']-df['Td']
    df=df.drop(['Td'],axis=1)
    # indexからhourを取り出しデータフレームhour（キー：'HOUR'）する．
    # 3時間ごとに時刻の区分（キー）をHOUR1, HOUR2, …, HOUR8に分ける．
    # 該当すれば1，該当しなければ0とする．後の学習の説明変数で使用する．
    # もう使用しないHOURは削除する．
    hour=pd.DataFrame({'HOUR': df.index.hour}, index=df.index)
    hour['HOUR1']=( ( hour['HOUR'] >= 0 ) & ( hour['HOUR'] < 3 ) ).astype(int)
    hour['HOUR2']=( ( hour['HOUR'] >= 3 ) & ( hour['HOUR'] < 6 ) ).astype(int)
    hour['HOUR3']=( ( hour['HOUR'] >= 6 ) & ( hour['HOUR'] < 9 ) ).astype(int)
    hour['HOUR4']=( ( hour['HOUR'] >= 9 ) & ( hour['HOUR'] < 12 ) ).astype(int)
    hour['HOUR5']=( ( hour['HOUR'] >= 12 ) & ( hour['HOUR'] < 15 ) ).astype(int)
    hour['HOUR6']=( ( hour['HOUR'] >= 15 ) & ( hour['HOUR'] < 18 ) ).astype(int)
    hour['HOUR7']=( ( hour['HOUR'] >= 18 ) & ( hour['HOUR'] < 21 ) ).astype(int)
    hour['HOUR8']=( ( hour['HOUR'] >= 21 ) & ( hour['HOUR'] < 24 ) ).astype(int)
    hour=hour.drop(['HOUR'],axis=1)
    #　データフレームdfにデータフレームhour列を追加する
    df=pd.merge(df, hour, how="outer", left_index=True, right_index=True)
    # indexからdayを取り出しデータフレームday（キー：'DAY'）とする．
    # 10日間ごとに日付けの区分（キー）をDAY1, DAY2, DAY3に分ける．
    # 該当すれば1，該当しなければ0とする．後の学習の説明変数とする．
    # もう使用しないDAYは削除する．
    day=pd.DataFrame({'DAY': df.index.day}, index=df.index)
    day['DAY1']=( ( day['DAY'] >= 1 ) & ( day['DAY'] < 11 ) ).astype(int)
    day['DAY2']=( ( day['DAY'] >= 11 ) & ( day['DAY'] < 21 ) ).astype(int)
    day['DAY3']=( ( day['DAY'] >= 21 ) & ( day['DAY'] < 32 ) ).astype(int)
    day=day.drop(['DAY'],axis=1)
    #　データフレームdfにデータフレームday列を追加する
    df=pd.merge(df, day, how="outer", left_index=True, right_index=True)
    # indexからweekdayを取り出しデータフレームweek（キー：'WEEK'）とする．
    # 土日（5,6）ならば1、平日（0,1,2,3,4,）ならば0とする。 
    week=pd.DataFrame({'WEEK': df.index.weekday}, index=df.index)
    week['WEEK']=( week['WEEK'] >= 5 ).astype(int)
    #　データフレームdfにデータフレームweek列を追加する
    df=pd.merge(df, week, how="outer", left_index=True, right_index=True)
    # indexからmonthを取り出しデータフレームmonth（キー：'MONTH'）とする．
    # 3ヶ月ごとに月の区分（キー）をDJF, MAM, JJA, SONに分ける．
    # 該当すれば1，該当しなければ0とする．後の学習の説明変数とする．
    # もう使用しないMONTHは削除する．
    month=pd.DataFrame({'MONTH': df.index.month}, index=df.index)
    month['DJF']=( ( month['MONTH'] == 12 ) | ( month['MONTH'] ==  1 ) | ( month['MONTH'] ==  2 ) ).astype(int)
    month['MAM']=( ( month['MONTH'] ==  3 ) | ( month['MONTH'] ==  4 ) | ( month['MONTH'] ==  5 ) ).astype(int)
    month['JJA']=( ( month['MONTH'] ==  6 ) | ( month['MONTH'] ==  7 ) | ( month['MONTH'] ==  8 ) ).astype(int)
    month['SON']=( ( month['MONTH'] ==  9 ) | ( month['MONTH'] == 10 ) | ( month['MONTH'] == 11 ) ).astype(int)
    month=month.drop(['MONTH'],axis=1)
    #　データフレームdfにデータフレームmonth列を追加する
    df=pd.merge(df, month, how="outer", left_index=True, right_index=True)
    # indexから'POWER'を取り出し24時間後にずらしてデータフレームpower24（キー：'POWER24'）とする．
    # 後の学習の説明変数とする． 
    power24=pd.DataFrame({'POWER24': df['POWER'].shift(+24)}, index=df.index)
    #　データフレームdfにデータフレームpower3列を追加する
    df=pd.merge(df, power24, how="outer", left_index=True, right_index=True)
    # indexから'POWER'を取り出し1時間後にずらしてデータフレームpower1（キー：'POWER1'）とする．
    # 後の学習の説明変数とする． 
    power1=pd.DataFrame({'POWER1': df['POWER'].shift(+1)}, index=df.index)
    #　データフレームdfにデータフレームpower1列を追加する
    df=pd.merge(df, power1, how="outer", left_index=True, right_index=True)
    # indexから'POWER'を取り出し2時間後にずらしてデータフレームpower2（キー：'POWER2'）とする．
    # 後の学習の説明変数とする． 
    power2=pd.DataFrame({'POWER2': df['POWER'].shift(+2)}, index=df.index)
    #　データフレームdfにデータフレームpower2列を追加する
    df=pd.merge(df, power2, how="outer", left_index=True, right_index=True)
    # indexから'POWER'を取り出し3時間後にずらしてデータフレームpower3（キー：'POWER3'）とする．
    # 後の学習の説明変数とする． 
    power3=pd.DataFrame({'POWER3': df['POWER'].shift(+3)}, index=df.index)
    #　データフレームdfにデータフレームpower3列を追加する
    df=pd.merge(df, power3, how="outer", left_index=True, right_index=True)
    # NaNがある行を取り除く 
    df=df.dropna(how='any')
    return df

#def 4.############################################
# データを学習用（train）と検証用（test）に別ける．
def preprocess(df,x_cols,y_cols,split):
    # データフレームdfからx_colsの列のデータの値（.values）を
    # float（浮動小数点型）として取り出す．
    x=df.loc[:,x_cols].values.astype('float')
    # データフレームdfからy_colsの列のデータの値（.values）を
    # float（浮動小数点型）として取り出す．
    # ravel()メソッドで1次元化する．
    y=df.loc[:,y_cols].values.astype('float').ravel()
    # 0からsplit-1までのxを学習用データとする
    x_train= x[:split]
    # 0からsplit-1までのyを学習用データとする．
    y_train= y[:split]
    # splitから終わりまでのxを検証用データとする．
    x_test= x[split:]
    # splitから終わりまでのyを検証用データとする．
    y_test= y[split:]
    return x_train,y_train,x_test,y_test

#def 5.############################################
# 予測モデルの結果（testおよびtrain）から時系列図を作成します
def timeseries(datetime_x, y1, y2, filename): 
    # dfのインデックス（時間）をXとする
    X=datetime_x
    # dfのname1列を指定してデータを取り出し，numpy配列で値をY1に与える．
    Y1=y1
    # dfのname1列を指定してデータを取り出し，numpy配列で値をY2に与える．
    Y2=y2
    # 時系列図の大きさを指定
    plt.figure(figsize=(20, 10))
    # y_obsvの時系列図
    plt.plot(X,Y1,color='blue',label='observed')
    # y_frcstの時系列図 
    plt.plot(X,Y2,color='red',label='predicted')
    # グラフのタイトル
    plt.title("Timeseries")
    # x軸のラベル
    plt.xlabel('Time')
    # y軸（左側の第1軸）のラベル
    plt.ylabel('Electric Power [$10^6$kW]')
    # 凡例（左上に置く） 
    plt.legend(loc='upper left')
    # 2つ目(name1)の凡例（右上に置く）
    # 保存するファイル名 
    plt.savefig(filename)
    # 図を閉じる
    plt.close()
    return

#def 6.############################################
# 予測モデルの結果（testおよびtrain）から散布図を作成します
def scatter(x, y,score, filename): 
    # 文字列"R2=score"
    score="R2="+str(score)
    # 散布図の大きさを指定
    plt.figure(figsize=(8, 8))
    # 散布図のプロット
    plt.plot(x, y, 'o')
    # Y=Xの線を引く
    plt.plot(x, x)
    # 文字列R2=r2)"を図の左上に置く
    plt.text(np.nanmin(x), np.nanmax(x), score)
    # グラフのタイトル
    plt.title("Scatter diagram")
    # x軸のラベル
    plt.xlabel('Observed Electric Power [$10^6$kW]')
    # y軸のラベル
    plt.ylabel('Predicted Electric Power [$10^6$kW]')    
    # 保存するファイル名 
    plt.savefig(filename)
    # 図を閉じる
    plt.close()
    return

###################################################
# ここから，メインプログラム
###################################################

#main 1.###########################################
# 気象庁AMeDAS（東京）の気温（℃）と露点温度（℃）のcsvファイルの名前を指定します．
filename1='amedas.csv'
# csvファイルの上から5行目まではデータではないため呼び飛ばします．
skipline1=5
# 気象庁AMeDAS（東京）の気温（℃）と露点温度（℃）のcsvファイルを読み込んで，
# pandasのDataFrame（tepco）に割り当てる関数を呼び出します．
amedas=readamedas(filename1,skipline1)
# DataFrame(amedas)の中のdummy1～dummy7の列を削除する．
amedas=amedas.drop(['dummy1','dummy2','dummy3','dummy4'],axis=1)
#amedas

#main 2.##########################################
# 東京電力の電力消費量（10^6 kW）のcsvファイルの名前を指定します．
filename2='tepco.csv'
# csvファイルの上から3行目まではデータではないため呼び飛ばします．
skipline2=3
# 東京電力の電力消費量（10^6 kW）のcsvファイルを読み込んで，
# pandasのデータフレーム（tepco）に割り当てる関数を呼び出します．
tepco=readtepco(filename2,skipline2)
# tepco

#main 3.##########################################
# 2つのデータフレーム（amedasとtepco）を結合して，
# 欠測値を除去し，気温以外に説明変数となりそうな情報を列に追加します．
data=dataprocess(amedas,tepco)

#main 4.##########################################
# データフレームdataの中から予測モデルの説明変数とするものを選ぶ
# さまざまな説明変数で試してみましょう．
x_cols=[]
# 気温
x_cols.extend(['T'])
# 湿数
#x_cols.extend(['T-Td'])
# 曜日
x_cols.extend(['WEEK'])
# 時間
x_cols.extend(['HOUR1','HOUR2','HOUR3','HOUR4','HOUR5','HOUR6','HOUR7','HOUR8'])
# 日にち
#x_cols.extend(['DAY1','DAY2','DAY3'])
# 月
#x_cols.extend(['DJF','MAM','JJA','SON'])
# 1日前の消費電力
#x_cols.extend(['POWER24'])
# 1時間前の消費電力
#x_cols.extend(['POWER1'])
# 2時間前の消費電力
#x_cols.extend(['POWER2'])
# 3時間前の消費電力
#x_cols.extend(['POWER3'])
# データフレームdataの中から予測モデルの目的変数とするものを選ぶ．
# ここでは発電量データ（'POWER'）を目的変数とする．
y_cols=['POWER']
# 学習用データと検証用データに別けるために，全体の行（len(data)）を2で割る
dt1=int(len(data)/2)
# データフレームdataを訓練データ（x_train，y_train）と
# 検証データ（x_test，y_test）に別ける
# dt1より前半（主に2017年）のデータを訓練用データとする．
# dt1より後半（主に2018年）のデータを検証用データとする．
# リストx_colsで設定した列が説明変数として扱われる．
# リストy_colsで設定した列が目的変数として扱われる．
x_train, y_train, x_test, y_test=preprocess(data, x_cols, y_cols, dt1)
#print('x_train.shape=',x_train.shape)
#print('y_train.shape=',y_train.shape)
#print('x_test.shape=',x_test.shape)
#print('y_test.shape=',y_test.shape)

#main 5.##########################################
# 説明変数に対して，平均0，標準偏差1となるような標準化を行う．
# preprocessing.StandardScaler()からscalerというインスタンスを作成する．
scaler=preprocessing.StandardScaler()
# fitメソッドにより説明変数x_trainの平均と分散を計算して記憶する．
scaler.fit(x_train)
# 説明変数x_trainに対して標準化を行い，変換後の配列を返す．
x_train=scaler.transform(x_train)
# 説明変数x_testに対して標準化を行い，変換後の配列を返す．
x_test=scaler.transform(x_test)

#main 6.##########################################
# modelインスタンスを作成
# ニューラルネットワークモデルの場合：
# 　solver="adam" 最適化手法（lbfgs, sgd, adam）
# 　activateion="relu" 活性化関数（identify, logistic, tanh, relu）
# 　max_iter=10000 反復の最大回数
# 　tol=1e-5　学習の収束値
# 　hidden_layer_sizes=(10,10,)　隠れ層のノード数（多層化可）
# 　alpha=0.0001 L2正則化のペナルティー係数
# 　batch_size='auto' バッチサイズの設定
# 　random_state=1 重み係数やバイアスの初期値やバッチサンプリングに用いられる乱数の設定
model = neural_network.MLPRegressor(solver="adam", activation="relu", hidden_layer_sizes=(10,10,), max_iter=10000, tol=1e-5, random_state=1, verbose=True)
# 線形重回帰モデルの場合：
#model = linear_model.LinearRegression()
# 線形サポートベクターマシン（線形カーネル）の場合：
#model = SVR(kernel='linear', C=1e3)
# 非線形サポートベクターマシン（ガウスカーネル）の場合：
# 　C=1e3　サポートベクターマシンと正則化項のバランス
# 　gamma=0.1　訓練データを中心としたガウス分布の広がり（分散の逆数）を表す
#model = SVR(kernel='rbf', C=1e3, gamma=0.1)
# fitメソッドにより目的変数y_train，説明変数x_trainによる
# 予測モデルの最適化
model.fit(x_train, y_train)
# 予測モデルのパラメータを取得する
#print(model.get_params(deep=True))
# 学習済の予測モデルを保存する． 
#with open('model.pickle', mode='wb') as fp:
#    pickle.dump(model, fp)
# 学習済の予測モデルを復元する．
#with open('model.pickle', mode='rb') as fp:
#    model=pickle.load(fp)
# 決定係数R2（学習データ）
score_train=model.score(x_train, y_train)
print('train R2 score=', score_train)
#　決定係数R2（検証データ）
score_test=model.score(x_test, y_test)
print('test R2 score=', score_test)
# 予測データの作成（学習データ）
y_train_predict=model.predict(x_train)
# 予測データの作成（検証データ）
y_test_predict=model.predict(x_test)
# 標準化したデータを元に戻す（学習データ）
x_train=scaler.inverse_transform([x_train])
# 標準化したデータを元に戻す（検証データ）
x_test=scaler.inverse_transform([x_test])

#main 7.##########################################
# 予測モデルの結果の可視化
# 予測モデルに使用するデータフレームdataの中のindex（時間情報）を
# datetime_xとする
datetime_x=list(data.index)
# 2018年1月1日のデータの行番号を取得する
# 学習用データと検証用データに別けるために，全体の行（len(data)）を2で割る
dt1=int(len(data)/2)
# 学習用データの1ヶ月分の行
dtm1=744
# 2018年2月1日のデータの行番号を取得する
dtm2=dt1+dtm1
# データを用いて時系列図を作成します．
# 訓練データの1年分の時系列図　ファイル名：timeseries_train.png
timeseries(datetime_x[:dt1],y_train,y_train_predict, "timeseries_train.png")
# 検証データの1年分の時系列図　ファイル名：timeseries_test.png
timeseries(datetime_x[dt1:],y_test,y_test_predict, "timeseries_test.png")
# 訓練データの1ヶ月分の時系列図　ファイル名：timeseries_train_jan.png
timeseries(datetime_x[:dtm1], y_train[:dtm1], y_train_predict[:dtm1], "timeseries_train_jan.png")
# 検証データの1ヶ月分の時系列図　ファイル名：timeseries_test_jan.png
timeseries(datetime_x[dt1:dtm2],y_test[:dtm1], y_test_predict[:dtm1], "timeseries_test_jan.png")
# データを用いて散布図を作成します．
# 訓練データの1年分の時系列　ファイル名：scatter_train.png
scatter(y_train, y_train_predict, score_train, "scatter_train.png")
# 検証データの1年分の時系列　ファイル名：scatter_test.png
scatter(y_test, y_test_predict, score_test, "scatter_test.png")